import json
import os
from datetime import datetime, timedelta

import requests
from bravado.exception import HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
from celery import shared_task

from django.conf import settings
from django.db import Error
from django.utils import timezone
from esi.errors import TokenError
from eveuniverse.models import EveMarketPrice

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from buybackprogram.app_settings import (
    BUYBACKPROGRAM_PRICE_INSTANT_PRICES,
    BUYBACKPROGRAM_PRICE_JANICE_API_KEY,
    BUYBACKPROGRAM_PRICE_METHOD,
    BUYBACKPROGRAM_PRICE_SOURCE_ID,
    BUYBACKPROGRAM_PRICE_SOURCE_NAME,
)
from buybackprogram.models import ItemPrices, Owner, Program, Tracking, TrackingItem

from .app_settings import (
    BUYBACKPROGRAM_TASKS_TIME_LIMIT,
    BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT,
)

logger = get_extension_logger(__name__)

NORMAL_TASK_PRIORITY = 4

# Create your tasks here
TASK_DEFAULT_KWARGS = {
    "time_limit": BUYBACKPROGRAM_TASKS_TIME_LIMIT,
}

TASK_ESI_KWARGS = {
    **TASK_DEFAULT_KWARGS,
    **{
        "bind": True,
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": 3},
        "retry_backoff": 30,
    },
}


def valid_janice_api_key():
    c = requests.get(
        "https://janice.e-351.com/api/rest/v2/markets",
        headers={
            "Content-Type": "text/plain",
            "X-ApiKey": BUYBACKPROGRAM_PRICE_JANICE_API_KEY,
            "accept": "application/json",
        },
    ).json()

    if "status" in c:
        logger.debug("Janice API status: %s" % c)
        return False
    else:
        return True


def get_bulk_prices(type_ids):
    r = None
    if BUYBACKPROGRAM_PRICE_METHOD == "Fuzzwork":
        r = requests.get(
            "https://market.fuzzwork.co.uk/aggregates/",
            params={
                "types": ",".join([str(x) for x in type_ids]),
                "station": BUYBACKPROGRAM_PRICE_SOURCE_ID,
            },
        ).json()
    elif BUYBACKPROGRAM_PRICE_METHOD == "Janice":
        r = requests.post(
            "https://janice.e-351.com/api/rest/v2/pricer?market=2",
            data="\n".join([str(x) for x in type_ids]),
            headers={
                "Content-Type": "text/plain",
                "X-ApiKey": BUYBACKPROGRAM_PRICE_JANICE_API_KEY,
                "accept": "application/json",
            },
        ).json()

        # Make Janice data look like Fuzzworks
        output = {}
        for item in r:
            output[str(item["itemType"]["eid"])] = {
                "buy": {
                    "max": str(item["immediatePrices"]["buyPrice5DayMedian"]),
                    "percentile": str(item["top5AveragePrices"]["buyPrice5DayMedian"]),
                },
                "sell": {
                    "min": str(item["immediatePrices"]["sellPrice5DayMedian"]),
                    "percentile": str(item["top5AveragePrices"]["sellPrice5DayMedian"]),
                },
            }
        r = output
    else:
        raise f"Unknown pricing method: {BUYBACKPROGRAM_PRICE_METHOD}"

    return r


@shared_task
def update_all_prices():
    type_ids = []
    market_data = {}
    api_up = True

    # Get all type ids
    prices = ItemPrices.objects.all()

    if BUYBACKPROGRAM_PRICE_METHOD == "Fuzzwork":
        logger.debug(
            "Price setup starting for %s items from Fuzzworks API from station id %s (%s), this may take up to 30 seconds..."
            % (
                len(prices),
                BUYBACKPROGRAM_PRICE_SOURCE_ID,
                BUYBACKPROGRAM_PRICE_SOURCE_NAME,
            )
        )
    elif BUYBACKPROGRAM_PRICE_METHOD == "Janice":
        if valid_janice_api_key():
            logger.debug(
                "Price setup starting for %s items from Janice API for Jita 4-4, this may take up to 30 seconds..."
                % (len(prices),)
            )
        else:
            logger.debug(
                "Price setup failed for Janice, invalid API key! Provide a working key or change price source to Fuzzwork"
            )
            api_up = False
    else:
        logger.error(
            "Unknown pricing method: '%s', skipping" % BUYBACKPROGRAM_PRICE_METHOD
        )
        return

    if api_up:
        # Build suitable bulks to fetch prices from API
        for item in prices:
            type_ids.append(item.eve_type_id)

            if len(type_ids) == 1000:
                market_data.update(get_bulk_prices(type_ids))
                type_ids.clear()

        # Get leftover data from the bulk
        if len(type_ids) > 0:
            market_data.update(get_bulk_prices(type_ids))

        logger.debug("Market data fetched, starting database update...")
        missing_items = []
        for price in prices:
            # Check if we received data from the API for the item. This will fix errors when using Janice as the POST endpoint does not return anything when there are no prices.
            if str(price.eve_type_id) in market_data:
                # Check what prices we should use either instant prices or top 5% percentile
                if not BUYBACKPROGRAM_PRICE_INSTANT_PRICES:
                    # Get the price values from the API data
                    buy = float(
                        market_data[str(price.eve_type_id)]["buy"]["percentile"]
                    )
                    sell = float(
                        market_data[str(price.eve_type_id)]["sell"]["percentile"]
                    )
                else:
                    # Get the price values from the API data
                    buy = float(market_data[str(price.eve_type_id)]["buy"]["max"])
                    sell = float(market_data[str(price.eve_type_id)]["sell"]["min"])

            # If API did not return any values we remove prices for the item
            else:
                missing_items.append(price.eve_type.name)

                # Reset prices for items not found from API
                buy = 0
                sell = 0

            price.buy = buy
            price.sell = sell
            price.updated = timezone.now()

        try:
            ItemPrices.objects.bulk_update(prices, ["buy", "sell", "updated"])
            logger.debug("All prices succesfully updated")
        except Error as e:
            logger.error("Error updating prices: %s" % e)

        EveMarketPrice.objects.update_from_esi()

        logger.debug("Updated all eveuniverse market prices.")

        if len(missing_items) > 0:
            logger.error(
                "%s items missing items from source API, prices set to 0."
                % len(missing_items)
            )

    else:
        logger.error("Price source API is not up! Prices not updated.")

    if BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT > 0:
        """cleanup unused tracking objects"""
        logger.debug(
            "Starting tracking objects cleanup. Removing tracking objects with no contracts assigned to them that are more than %s hours old "
            % BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT
        )

        try:
            trackings, t = Tracking.objects.filter(
                contract_id__isnull=True,
                created_at__lte=timezone.now()
                - timedelta(hours=BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT),
            ).delete()
            logger.debug("%s old unlinked tracking objects deleted." % len(t))
        except Error as e:
            logger.error("Error deleting old tracking objects: %s" % e)
    else:
        logger.debug(
            "Tracking object time limit is set to %s, no old tracking object purging will happen. "
            % BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT
        )


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["owner_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_contracts_for_owner(self, owner_pk):
    """fetches all contracts for owner from ESI"""

    try:
        return _get_owner(owner_pk).update_contracts_esi()
    except TokenError:
        print("Invalid token provided.")
        return None
    except Exception as e:
        print(f"An unexpected error occurred: {str(e)}")
        return None


@shared_task(**TASK_DEFAULT_KWARGS)
def update_all_contracts():
    logger.debug("Starting all contract updates")
    for owner in Owner.objects.all():
        logger.debug("Updating contracts for %s" % owner)
        update_contracts_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk},
            priority=NORMAL_TASK_PRIORITY,
        )


def _get_owner(owner_pk: int) -> Owner:
    """returns the owner or raises exception"""
    try:
        owner = Owner.objects.get(pk=owner_pk)
    except Owner.DoesNotExist:
        raise Owner.DoesNotExist(
            "Requested owner with pk {} does not exist".format(owner_pk)
        )
    return owner


@shared_task(**TASK_DEFAULT_KWARGS)
def update_program_performance():
    programs = Program.objects.all()
    logger.debug("Got programs %s" % programs)
    for program in programs:
        logger.debug("Starting performance calculations for %s" % program.pk)
        firstbench = datetime.now()
        lastbench = datetime.now()
        # Tracker values
        monthstats = {
            "status": {},
            "overall": {"all": {}},
            "items": {},
            "categories": {},
            "donations": {"all": {}},
        }
        allmonths = set()

        # Category to Item mapping
        category2items = {}

        # Export data
        dumpdata = [
            [
                "Contract ID",
                "Date Issued",
                "Date Finished",
                "User ID",
                "Total ISK",
                "Object Cateogry",
                "Object ID",
                "Object Name",
                "Object Quant",
                "Object ISK",
            ]
        ]

        # Get all tracking objects that have a linked contract to them for the user
        tracking_numbers = (
            Tracking.objects.filter(program_id=program.pk)
            .filter(contract__isnull=False)
            .prefetch_related("contract")
        )

        logger.debug(
            "Performance bench: %.2f" % (datetime.now() - lastbench).total_seconds()
        )
        lastbench = datetime.now()
        # Loop all tracking objects
        for tracking in tracking_numbers:
            month = datetime.strftime(tracking.contract.date_issued, "%Y-%m") + "-15"
            allmonths.add(month)

            if month not in monthstats["status"]:
                monthstats["status"][
                    month
                ] = {}  # status of all contracts issued during a given month

            # Gather stats on all contracts' statuses
            if tracking.contract.status not in monthstats["status"][month]:
                monthstats["status"][month][tracking.contract.status] = 0
            monthstats["status"][month][tracking.contract.status] += 1

            # For finished contracts, gather more data
            if tracking.contract.status == "finished":
                # "overall": {"all": {"isk": {}, "q": {}, }
                if month not in monthstats["overall"]["all"]:
                    monthstats["overall"]["all"][month] = [0, 0]  # isk, q
                    monthstats["donations"]["all"][month] = [0, 0]  # isk, q

                monthstats["overall"]["all"][month][0] += tracking.contract.price
                monthstats["overall"]["all"][month][1] += 1

                monthstats["donations"]["all"][month][0] += tracking.donation
                if tracking.donation > 0:
                    monthstats["donations"]["all"][month][1] += 1

                logger.debug(
                    "Performance loop bench: %.2f"
                    % (datetime.now() - lastbench).total_seconds()
                )
                lastbench = datetime.now()
                # Collect ISK data per items
                tracking_items = TrackingItem.objects.filter(tracking=tracking)
                for item in tracking_items:
                    if item.eve_type.name not in monthstats["items"]:
                        monthstats["items"][item.eve_type.name] = {}
                    if month not in monthstats["items"][item.eve_type.name]:
                        monthstats["items"][item.eve_type.name][month] = [0, 0]
                    monthstats["items"][item.eve_type.name][month][0] += item.buy_value
                    monthstats["items"][item.eve_type.name][month][1] += item.quantity
                    # cache[tracking.contract.issuer_id] = EveEntity.objects.resolve_name(tracking.contract.issuer_id)

                    # Collect item category data
                    catid = item.eve_type.eve_group.name
                    if catid not in category2items:
                        category2items[catid] = set()
                    category2items[catid].add(item.eve_type.name)

                    if catid not in monthstats["categories"]:
                        monthstats["categories"][catid] = {}
                    if month not in monthstats["categories"][catid]:
                        monthstats["categories"][catid][month] = [0, 0]
                    monthstats["categories"][catid][month][0] += item.buy_value
                    monthstats["categories"][catid][month][1] += item.quantity

                    logger.debug("Processing contract %s" % tracking.contract.id)

                    # Collect data for export
                    dumpdata.append(
                        [
                            tracking.id,
                            datetime.strftime(
                                tracking.contract.date_issued, "%Y-%m-%d %H:%M:%S"
                            ),
                            datetime.strftime(
                                tracking.contract.date_completed, "%Y-%m-%d %H:%M:%S"
                            ),
                            tracking.contract.issuer_id,
                            tracking.contract.price,
                            item.eve_type.eve_group.name,
                            item.eve_type.id,
                            item.eve_type.name,
                            item.quantity,
                            item.buy_value,
                        ]
                    )

        logger.debug(
            "Performance exit loop bench: %.2f"
            % (datetime.now() - lastbench).total_seconds()
        )
        lastbench = datetime.now()
        # Reformat data so that it is easier to use billboard.js
        allmonths = sorted(list(allmonths))

        scaling = {}
        hscaling = {}
        for strata in ("overall", "items", "categories", "donations"):
            scaling[strata] = []
            for yi in monthstats[strata].keys():
                scaling[strata] += [
                    monthstats[strata][yi][x][0] for x in monthstats[strata][yi].keys()
                ]
            if len(scaling[strata]) == 0:
                scaling[strata] = 1
            else:
                scaling[strata] = sum(scaling[strata]) / len(scaling[strata])
            for s, h in ((1e9, "Billions"), (1e6, "Millions")):
                if scaling[strata] > s:
                    scaling[strata] = s
                    hscaling[strata] = h
                    break
            if scaling[strata] < 1e6:
                scaling[strata] = 1
                hscaling[strata] = ""

        # Always set donations to be the same scale as overall since they are plotted together.
        scaling["donations"] = scaling["overall"]
        hscaling["donations"] = hscaling["overall"]

        for strata in ("overall", "items", "categories", "donations"):
            for yi in monthstats[strata].keys():
                y = [[yi], [yi]]
                # Overall ISK
                if strata == "overall":
                    y = [["Bought"], ["Contract count"]]
                # Donation ISK
                if strata == "donations":
                    y = [["Donations"], ["Donation count"]]

                for m in allmonths:
                    if m not in (monthstats[strata][yi]):
                        monthstats[strata][yi][m] = [0, 0]
                    y[0].append(
                        round(float(monthstats[strata][yi][m][0]) / scaling[strata], 3)
                    )
                    y[1].append(float(monthstats[strata][yi][m][1]))
                monthstats[strata][yi] = y
        monthstats["x"] = ["x"] + allmonths

        for cat in category2items:
            category2items[cat] = list(category2items[cat])

        # Break down of categories by last three months:
        lastthree = {}
        for yi in monthstats["categories"].keys():
            calc = monthstats["categories"][yi][0].copy()
            calc.pop(0)
            lastthree[yi] = sum(calc[-3:])

        if len(lastthree.keys()) > 10:
            th = lastthree[sorted(lastthree.keys(), key=lambda x: -lastthree[x])[10]]
            lastthree["Other"] = 0
            for k in list(lastthree.keys()):
                if k != "Other" and lastthree[k] <= th:
                    lastthree["Other"] += lastthree[k]
                    del lastthree[k]

            monthstats["categories"]["Other"] = None
            for k in list(monthstats["categories"].keys()):
                if k not in lastthree:
                    if monthstats["categories"]["Other"] is None:
                        monthstats["categories"]["Other"] = monthstats["categories"][k]
                    else:
                        monthstats["categories"]["Other"][0][1:] = [
                            sum(x)
                            for x in zip(
                                monthstats["categories"]["Other"][0][1:],
                                monthstats["categories"][k][0][1:],
                            )
                        ]
                        monthstats["categories"]["Other"][1][1:] = [
                            sum(x)
                            for x in zip(
                                monthstats["categories"]["Other"][1][1:],
                                monthstats["categories"][k][1][1:],
                            )
                        ]
                    del monthstats["categories"][k]
            monthstats["categories"]["Other"][0][0] = "Other"
            monthstats["categories"]["Other"][1][0] = "Other"

        lastthree = list(lastthree.items())

        # Sanitize CSV data
        for i in range(len(dumpdata)):
            for j in range(len(dumpdata[i])):
                dumpdata[i][j] = str(dumpdata[i][j]).replace(",", " ")

        context = {
            "stats": json.dumps(monthstats),
            "lastthree": json.dumps(lastthree),
            "categories": json.dumps(category2items),
            "export": json.dumps(dumpdata),
            "hscaling": json.dumps(hscaling),
        }
        logger.debug(
            "Performance finished: %.2f" % (datetime.now() - lastbench).total_seconds()
        )
        logger.debug(
            "Performance total: %.2f" % (datetime.now() - firstbench).total_seconds()
        )

        logger.debug(context)

        static_path = os.path.join(
            settings.STATIC_ROOT, "buybackprogram/performance_data"
        )
        filename = f"program_performance_{program.pk}.json"
        file_path = os.path.join(static_path, filename)

        data = context

        if not os.path.exists(static_path):
            os.makedirs(static_path)
        with open(file_path, "w") as file:
            json.dump(data, file)



from django import template

register = template.Library()


@register.filter
def price(value):
    if value:
        return value + " ISK"
    else:
        return "-"


@register.filter
def tax(value):
    if value:
        return str(value) + " %"
    else:
        return "-"


@register.filter
def comparison(value):
    if float(value) > 0:
        return "+ " + str(value) + " %"
    else:
        return str(value) + " %"


@register.filter
def custom_number_format(value, decimal_places=2):
    try:
        # Convert the value to a float
        value = float(value)
        # Format with a space as the thousand separator and a period as the decimal separator
        formatted_number = f"{value:,.{decimal_places}f}".replace(",", " ").replace(
            ".", "."
        )
        return formatted_number
    except (ValueError, TypeError):
        return ""



from datetime import timedelta
from random import choice, randint

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.utils import timezone

from allianceauth.eveonline.models import EveCorporationInfo

from buybackprogram.models import Contract, EveType, Program, Tracking, TrackingItem

User = get_user_model()


class Command(BaseCommand):
    help = "Generates dummy data for contracts, tracking objects, and tracking items. Only for development use"

    def add_arguments(self, parser):
        parser.add_argument(
            "-n",
            "--number",
            type=int,
            default=1000,
            help="Number of contracts and tracking objects to create",
        )

    def handle(self, *args, **options):
        # Prompt the user for confirmation
        self.stdout.write(
            self.style.WARNING(
                "This command will delete all existing contracts and tracking objects! This command is only for development use!"
            )
        )
        if input("Type 'yes' to continue, or 'no' to cancel: ").lower() != "yes":
            self.stdout.write(self.style.ERROR("Command aborted by user."))
            return

        # Delete all existing Contract and Tracking objects
        self.stdout.write(
            self.style.WARNING(
                "Deleting all existing contracts and tracking objects..."
            )
        )
        Contract.objects.all().delete()
        Tracking.objects.all().delete()

        self.stdout.write(
            self.style.SUCCESS(
                "All existing contracts and tracking objects have been deleted."
            )
        )
        number = options["number"]

        self.stdout.write(
            self.style.SUCCESS(
                f"Starting generation of {number} dummy contracts, tracking objects, and tracking items"
            )
        )

        for _ in range(number):
            user = User.objects.order_by("?").first()
            corporation = EveCorporationInfo.objects.order_by("?").first()
            program = Program.objects.order_by("?").first()

            days_ago = randint(0, 180)  # Random number of days up to 6 months
            date_issued = timezone.now() - timedelta(days=days_ago)
            date_expired = date_issued + timedelta(
                days=90
            )  # Contract expires 90 days after issue

            # Randomly selecting contract status
            status = choice(["outstanding", "finished", "in_progress"])
            date_completed = None

            # Set date_completed if the status is 'finished'
            if status == "finished":
                completion_days = randint(0, (date_expired - date_issued).days)
                date_completed = date_issued + timedelta(days=completion_days)

            contract = Contract.objects.create(
                assignee_id=randint(1000000, 9999999),
                availability="public",
                contract_id=randint(100000000, 999999999),
                date_issued=date_issued,
                date_expired=date_expired,
                date_completed=date_completed,
                for_corporation=choice([True, False]),
                issuer_corporation_id=corporation.corporation_id,
                issuer_id=1,
                start_location_id=randint(10000000, 99999999),
                price=randint(1000000, 100000000),
                status=status,
                volume=randint(100, 10000),
            )

            tracking = Tracking.objects.create(
                program=program,
                contract=contract,
                issuer_user=user,
                value=randint(500000, 50000000),
                taxes=randint(50000, 500000),
                hauling_cost=randint(10000, 200000),
                donation=randint(0, 10000),
                net_price=contract.price - randint(10000, 100000),
                tracking_number=f"TRACK-{randint(1000, 9999)}",
            )

            # Generate a few TrackingItem instances for each Tracking
            for _ in range(randint(1, 5)):  # Random number of items from 1 to 5
                eve_type = EveType.objects.order_by(
                    "?"
                ).first()  # Randomly select an item type
                if eve_type:
                    TrackingItem.objects.create(
                        tracking=tracking,
                        eve_type=eve_type,
                        quantity=randint(1, 100),
                        buy_value=randint(100, 10000),
                    )

            self.stdout.write(
                self.style.SUCCESS(
                    f"Generated contract {contract.contract_id}, tracking {tracking.id}, and tracking items"
                )
            )

        self.stdout.write(
            self.style.SUCCESS(
                f"Finished generating {number} dummy contracts, tracking objects, and tracking items"
            )
        )



from django import template

from buybackprogram.app_settings import BUYBACKPROGRAM_PRICE_INSTANT_PRICES

register = template.Library()


@register.simple_tag
def help(option):
    if not BUYBACKPROGRAM_PRICE_INSTANT_PRICES:
        source = "average top 5 %"
    else:
        source = "instant"

    icons = {
        "item_details": "Details for each of the items on this contract",
        "accepted_locations": "Locations where the contracts are accepted at",
        "contract_settings": "Settings used in the actual ingame contract",
        "invoice": "List of expenses applied on your contract",
        "name": "The type name for this item",
        "quantity": "Quantity of items to be sold",
        "price_source": "The buy and sell prices for this item at the selected trade hub",
        "base_price": "The bace price we use for our calculations before any taxes or expenses",
        "taxes": "Our taxes applied over the base price for this item",
        "price": "Our net price that we will pay for one unit of this item",
        "total": "Our net price for all of the units sold for this item",
        "notes": "Any notes for this item row will be displayed in here",
        "price_type": "Prices are based on %s prices" % source,
    }

    return icons[option]



# Includes all asteroids, moon ore and ice. All items that can be compressed or refined for minerals*
ORE_EVE_GROUPS = [
    450,  # Arkonor
    451,  # Bistot
    452,  # Crokite
    453,  # Ochre
    454,  # Hedbergite
    455,  # Hemorphite
    456,  # Jaspet
    457,  # Kernite
    458,  # Plagioclase
    459,  # Pyroxeres
    460,  # Scordite
    461,  # Spodumain
    462,  # Veldspar
    465,  # Ice
    467,  # Gneiss
    468,  # Mercoxit
    469,  # Omber
    1884,  # Zeolites
    1911,  # Raspite
    1920,  # Cobaltite
    1921,  # Otavite
    1922,  # Carnotite
    1923,  # Xenotime
    2024,  # Amethystic Crystallite
    2029,  # ??
    4029,  # Talassonite
    4030,  # Rakovene
    4031,  # Bezdnacine
    4094,  # ??
    4513,  # Mordunium
    4514,  # Ytirium
    4515,  # Eifyrium
    4516,  # Ducinium
    4759,  # Griemeer
    4758,  # Hezorime
    4755,  # Kylixium
    4756,  # Nocxite
    4757,  # Ueganite
    4915,  # Prismaticite
]

OPE_EVE_GROUPS = [493]

BONDS_EVE_GROUPS = [1248]

BLUE_LOOT_TYPE_IDS = [
    30744,
    30745,
    30746,
    30747,
]

RED_LOOT_TYPE_IDS = [
    48121,
    60459,
]



# Generated by Django 4.2.11 on 2024-11-03 04:42

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0016_alter_program_compression_price_dencity_modifier"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contract",
            name="price",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="contract",
            name="volume",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="itemprices",
            name="buy",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="itemprices",
            name="sell",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="tracking",
            name="donation",
            field=models.DecimalField(
                blank=True, decimal_places=2, max_digits=20, null=True
            ),
        ),
        migrations.AlterField(
            model_name="tracking",
            name="hauling_cost",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="tracking",
            name="net_price",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="tracking",
            name="taxes",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="tracking",
            name="value",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
        migrations.AlterField(
            model_name="trackingitem",
            name="buy_value",
            field=models.DecimalField(decimal_places=2, max_digits=20),
        ),
    ]



# Generated by Django 4.0.10 on 2023-12-31 13:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0013_program_discord_show_item_list"),
    ]

    operations = [
        migrations.AddField(
            model_name="program",
            name="price_type",
            field=models.CharField(
                choices=[("Buy", "Buy"), ("Sell", "Sell"), ("Split", "Split")],
                default="Buy",
                help_text="What prices should we use as the source for prices. Default: Buy",
                max_length=7,
            ),
        ),
    ]



from django import template

register = template.Library()


@register.filter
def BsTabColorClass(color):
    if color == "red":
        return "bg-danger"
    elif color == "orange":
        return "bg-warning"
    elif color == "green":
        return "bg-success"
    else:
        return "bg-primary"



# Generated by Django 4.2.11 on 2024-10-02 03:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0015_faq_contractnotification_header_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="program",
            name="compression_price_dencity_modifier",
            field=models.BooleanField(
                default=False,
                help_text="Should we use compressed volume for items that can be compressed such as ore and ice in fuel cost and price dencity calculations. If set to False the raw volume is used for volume calculations if the item does not have a compressed price variant available from the use compressed value setting.",
                verbose_name="Ore volume based on compressed volume",
            ),
        ),
    ]



# Generated by Django 4.2.11 on 2024-05-11 11:42

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0014_program_price_type"),
    ]

    operations = [
        migrations.CreateModel(
            name="Faq",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("header", models.CharField(max_length=1024)),
                ("body", models.TextField()),
            ],
        ),
        migrations.AddField(
            model_name="contractnotification",
            name="header",
            field=models.CharField(blank=True, max_length=1024, null=True),
        ),
        migrations.AddField(
            model_name="tracking",
            name="additional_notes",
            field=models.TextField(blank=True, null=True),
        ),
    ]



import json
import os
from datetime import datetime

from django.conf import settings
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.shortcuts import render
from django.utils import timezone
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

from buybackprogram.app_settings import BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS
from buybackprogram.notes import (
    note_missing_from_contract,
    note_missing_from_tracking,
    note_quantity_missing_from_contract,
    note_quantity_missing_from_tracking,
)
from buybackprogram.utils import messages_plus

from ..models import (
    Contract,
    ContractItem,
    ContractNotification,
    Program,
    Tracking,
    TrackingItem,
)

logger = get_extension_logger(__name__)


@login_required
@permission_required("buybackprogram.basic_access")
def my_stats(request):
    # List for valid contracts to be displayed
    valid_contracts = []
    contract_notes = []

    # Tracker values
    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    # Request user owned characters
    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(contract__isnull=False)
        .filter(contract__issuer_id__in=characters)
        .filter(contract__date_expired__gte=timezone.now())
        .prefetch_related("contract")
    )

    # Loop tracking objects to see if we have any contracts
    for tracking in tracking_numbers:
        # Get notes for this contract
        tracking.contract.notes = ContractNotification.objects.filter(
            contract=tracking.contract
        )

        # Walk the tracker values for contracts
        if tracking.contract.status == "outstanding":
            values["outstanding"] += tracking.contract.price
            values["outstanding_count"] += 1
        if tracking.contract.status == "finished":
            values["finished"] += tracking.contract.price
            values["finished_count"] += 1

        # Get the name for the issuer
        tracking.contract.issuer_name = EveEntity.objects.resolve_name(
            tracking.contract.issuer_id
        )

        # Get the name for the assignee
        tracking.contract.assignee_name = EveEntity.objects.resolve_name(
            tracking.contract.assignee_id
        )

        # Add contract to the valid contract list
        valid_contracts.append(tracking)
        contract_notes.append(tracking.contract.notes)

    context = {
        "contracts": valid_contracts,
        "contract_notes": contract_notes,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def leaderboard(request, program_pk):
    program = Program.objects.get(pk=program_pk)

    # Tracker values
    monthstats = {
        "users": {},  # monthly stats per user
        "userinfo": {},  # profile information per user
        "months": None,  # all months
    }
    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(program_id=program_pk)
        .filter(contract__isnull=False)
        .prefetch_related("contract")
    )

    # Loop all tracking objects
    for tracking in tracking_numbers:
        # For finished contracts, gather more data
        if tracking.contract.status == "finished":
            month = datetime.strftime(tracking.contract.date_issued, "%Y-%m:%B %Y")
            if month not in monthstats["users"]:
                monthstats["users"][month] = {}  # User data per month

            # Collect ISK data per user
            user = tracking.contract.issuer_id
            if user not in monthstats["users"][month]:
                monthstats["users"][month][user] = [
                    0,  # contract total
                    0,  # donation total
                ]
            monthstats["users"][month][user][0] += float(tracking.contract.price)
            monthstats["users"][month][user][1] += float(tracking.donation)

            if user not in monthstats["userinfo"]:
                monthstats["userinfo"][user] = {
                    "name": EveEntity.objects.resolve_name(user),
                    "pic": f"https://images.evetech.net/characters/{user}/portrait?size=32",
                }

    monthstats["months"] = sorted(list(monthstats["users"].keys()))
    context = {
        "stats": json.dumps(monthstats),
        "program": program,
    }

    return render(request, "buybackprogram/leaderboards.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_performance(request, program_pk):
    static_path = os.path.join(settings.STATIC_ROOT, "buybackprogram/performance_data")
    filename = f"program_performance_{program_pk}.json"
    file_path = os.path.join(static_path, filename)

    try:
        # Try to open the file and load the JSON data
        with open(file_path, "r") as file:
            context = json.load(file)
    except FileNotFoundError:
        messages_plus.error(
            request,
            "Error: The performance data file is missing for this program. Please generate the report first.",
        )
        return render(
            request,
            "buybackprogram/performance.html",
        )

    return render(request, "buybackprogram/performance.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_stats(request):
    # List for valid contracts to be displayed
    valid_contracts = []
    untracked_contracts = False

    # Tracker values
    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
        "untracked_count": 0,
    }

    # Request user owned characters
    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    # Request user owned corporations
    corporations = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__corporation_id", flat=True
    )

    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(contract__isnull=False)
        .filter(
            Q(contract__assignee_id__in=characters)
            | Q(contract__assignee_id__in=corporations)
        )
        .filter(contract__date_expired__gte=timezone.now())
        .prefetch_related("contract")
    )

    # Loop tracking objects to see if we have any contracts
    for tracking in tracking_numbers:
        # Get notes for this contract
        tracking.contract.notes = ContractNotification.objects.filter(
            contract=tracking.contract
        )

        # Walk the tracker values for contracts
        if tracking.contract.status == "outstanding":
            values["outstanding"] += tracking.contract.price
            values["outstanding_count"] += 1
        if tracking.contract.status == "finished":
            values["finished"] += tracking.contract.price
            values["finished_count"] += 1

        # Get the name for the issuer
        tracking.contract.issuer_name = EveEntity.objects.resolve_name(
            tracking.contract.issuer_id
        )

        # Get the name for the assignee
        tracking.contract.assignee_name = EveEntity.objects.resolve_name(
            tracking.contract.assignee_id
        )

        # Add contract to the valid contract list
        valid_contracts.append(tracking)

    if BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS:
        # Get pending contracts that have no tracking assigned to them
        untracked_contracts = Contract.objects.filter(
            Q(assignee_id__in=characters) | Q(assignee_id__in=corporations)
        ).filter(no_tracking=True, status="outstanding")

        logger.debug("Got %s untracked contracts" % len(untracked_contracts))

        for contract in untracked_contracts:
            values["untracked_count"] += 1

            # Get notes for this contract
            contract.notes = ContractNotification.objects.filter(contract=contract)

            # Get the name for the issuer
            contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

            # Get the name for the assignee
            contract.assignee_name = EveEntity.objects.resolve_name(
                contract.assignee_id
            )

    context = {
        "contracts": valid_contracts,
        "untracked_contracts": untracked_contracts,
        "values": values,
        "mine": True,
        "BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS": BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.see_all_statics")
def program_stats_all(request):
    # List for valid contracts to be displayed
    valid_contracts = []

    untracked_contracts = False

    # Tracker values
    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
        "untracked_count": 0,
    }

    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(contract__isnull=False)
        .filter(contract__date_expired__gte=timezone.now())
        .prefetch_related("contract")
    )

    # Loop tracking objects to see if we have any contracts
    for tracking in tracking_numbers:
        # Get notes for this contract
        tracking.contract.notes = ContractNotification.objects.filter(
            contract=tracking.contract
        )

        # Walk the tracker values for contracts
        if tracking.contract.status == "outstanding":
            values["outstanding"] += tracking.contract.price
            values["outstanding_count"] += 1
        if tracking.contract.status == "finished":
            values["finished"] += tracking.contract.price
            values["finished_count"] += 1

        # Get the name for the issuer
        tracking.contract.issuer_name = EveEntity.objects.resolve_name(
            tracking.contract.issuer_id
        )

        # Get the name for the assignee
        tracking.contract.assignee_name = EveEntity.objects.resolve_name(
            tracking.contract.assignee_id
        )

        valid_contracts.append(tracking)

    if BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS:
        # Get pending contracts that have no tracking assigned to them
        untracked_contracts = Contract.objects.filter(
            no_tracking=True, status="outstanding"
        )

        logger.debug("Got %s untracked contracts" % len(untracked_contracts))

        for contract in untracked_contracts:
            values["untracked_count"] += 1

            # Get notes for this contract
            contract.notes = ContractNotification.objects.filter(contract=contract)

            # Get the name for the issuer
            contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

            # Get the name for the assignee
            contract.assignee_name = EveEntity.objects.resolve_name(
                contract.assignee_id
            )

    context = {
        "contracts": valid_contracts,
        "untracked_contracts": untracked_contracts,
        "values": values,
        "mine": True,
        "BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS": BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def contract_details(request, contract_title):
    contract = Contract.objects.get(title__contains=contract_title)

    # Get notes for this contract
    notes = ContractNotification.objects.filter(contract=contract)

    # Get items for this contract
    contract_items = ContractItem.objects.filter(contract=contract)

    # Get tracking object for this contract
    tracking = Tracking.objects.get(
        tracking_number=contract_title,
    )

    # Get tracked items
    tracking_items = TrackingItem.objects.filter(tracking=tracking)

    # Find the difference in the created contract and original calculation
    for tracking_item in tracking_items:
        tracking_notes = []

        item_match = False
        quantity_match = False

        for contract_item in contract_items:
            if contract_item.eve_type == tracking_item.eve_type:
                item_match = True

                if contract_item.quantity == tracking_item.quantity:
                    quantity_match = True
                    break

        tracking_item.item_match = item_match

        if not item_match:
            tracking_notes.append(note_missing_from_contract(tracking_item.eve_type))

        if item_match and not quantity_match:
            tracking_notes.append(
                note_quantity_missing_from_contract(tracking_item.eve_type)
            )

        tracking_item.notes = tracking_notes

    for contract_item in contract_items:
        contract_notes = []

        item_match = False
        quantity_match = False

        for tracking_item in tracking_items:
            if contract_item.eve_type == tracking_item.eve_type:
                item_match = True

                if contract_item.quantity == tracking_item.quantity:
                    quantity_match = True
                    break

        contract_item.item_match = item_match

        if not item_match:
            contract_notes.append(note_missing_from_tracking(contract_item.eve_type))

        if item_match and not quantity_match:
            contract_notes.append(
                note_quantity_missing_from_tracking(contract_item.eve_type)
            )

        contract_item.notes = contract_notes

    # Get the name for the issuer
    contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

    # Get the name for the assignee
    contract.assignee_name = EveEntity.objects.resolve_name(contract.assignee_id)

    # Sort lists by reverse quantity order
    contract_items = sorted(contract_items, key=lambda x: -x.quantity)
    tracking_items = sorted(tracking_items, key=lambda x: -x.quantity)

    context = {
        "notes": notes,
        "contract": contract,
        "contract_items": contract_items,
        "tracking": tracking,
        "tracking_items": tracking_items,
    }

    return render(request, "buybackprogram/contract_details.html", context)
